@@ -6,19 +6,20 @@ import os
 import re
 import sys
 from collections import defaultdict
+from dataclasses import field
 from os import environ as env
 from os.path import dirname
-from typing import Any, Callable, Dict, List, Optional, Type, Union
+from typing import Any, Callable, Dict, List, Optional, Type, Union, cast
 from urllib.parse import urlparse
 
-from pydantic import validator
+from pydantic import Field, validator
 from pydantic.dataclasses import dataclass
 from pydantic.json import pydantic_encoder
 from ruamel.yaml import YAML
 from typing_extensions import Literal
 
 from dipdup.exceptions import ConfigurationError
-from dipdup.models import IndexType, State
+from dipdup.models import State
 from dipdup.utils import camel_to_snake, reindex, snake_to_camel
 
 ROLLBACK_HANDLER = 'on_rollback'
@@ -163,7 +164,7 @@ class OperationHandlerPatternConfig:
     @property
     def parameter_type_cls(self) -> Type:
         if self._parameter_type_cls is None:
-            raise Exception('Parameter type is not registered')
+            raise RuntimeError('Config is not initialized')
         return self._parameter_type_cls
 
     @parameter_type_cls.setter
@@ -173,7 +174,7 @@ class OperationHandlerPatternConfig:
     @property
     def storage_type_cls(self) -> Type:
         if self._storage_type_cls is None:
-            raise Exception('Storage type is not registered')
+            raise RuntimeError('Config is not initialized')
         return self._storage_type_cls
 
     @storage_type_cls.setter
@@ -182,15 +183,8 @@ class OperationHandlerPatternConfig:
 
 
 @dataclass
-class OperationHandlerConfig:
-    """Operation handler config
-
-    :param callback: Name of method in `handlers` package
-    :param pattern: Filters to match operations in group
-    """
-
+class HandlerConfig:
     callback: str
-    pattern: List[OperationHandlerPatternConfig]
 
     def __post_init_post_parse__(self):
         self._callback_fn = None
@@ -207,26 +201,22 @@ class OperationHandlerConfig:
 
 
 @dataclass
-class OperationIndexConfig:
-    """Operation index config
+class OperationHandlerConfig(HandlerConfig):
+    """Operation handler config
 
-    :param datasource: Alias of datasource in `datasources` block
-    :param contract: Alias of contract to fetch operations for
-    :param first_block: First block to process
-    :param last_block: Last block to process
-    :param handlers: List of indexer handlers
+    :param callback: Name of method in `handlers` package
+    :param pattern: Filters to match operations in group
     """
 
-    kind: Literal["operation"]
+    pattern: List[OperationHandlerPatternConfig]
+
+
+@dataclass
+class IndexConfig:
     datasource: Union[str, TzktDatasourceConfig]
-    contract: Union[str, ContractConfig]
-    handlers: List[OperationHandlerConfig]
-    first_block: int = 0
-    last_block: int = 0
 
     def __post_init_post_parse__(self):
         self._state: Optional[State] = None
-        self._rollback_fn: Optional[Callable] = None
         self._template_values: Dict[str, str] = None
 
     def hash(self) -> str:
@@ -238,16 +228,11 @@ class OperationIndexConfig:
         ).hexdigest()
 
     @property
-    def tzkt_config(self) -> TzktDatasourceConfig:
+    def datasource_config(self) -> TzktDatasourceConfig:
         if not isinstance(self.datasource, TzktDatasourceConfig):
             raise RuntimeError('Config is not initialized')
         return self.datasource
 
-    @property
-    def contract_config(self) -> ContractConfig:
-        assert isinstance(self.contract, ContractConfig)
-        return self.contract
-
     @property
     def state(self):
         if not self._state:
@@ -258,16 +243,6 @@ class OperationIndexConfig:
     def state(self, value: State):
         self._state = value
 
-    @property
-    def rollback_fn(self) -> Callable:
-        if not self._rollback_fn:
-            raise RuntimeError('Config is not initialized')
-        return self._rollback_fn
-
-    @rollback_fn.setter
-    def rollback_fn(self, value: Callable) -> None:
-        self._rollback_fn = value
-
     @property
     def template_values(self) -> Optional[Dict[str, str]]:
         return self._template_values
@@ -278,58 +253,103 @@ class OperationIndexConfig:
 
 
 @dataclass
-class BigmapdiffHandlerPatternConfig:
-    name: str
-    entry_type: str
+class OperationIndexConfig(IndexConfig):
+    """Operation index config
 
+    :param datasource: Alias of datasource in `datasources` block
+    :param contract: Alias of contract to fetch operations for
+    :param first_block: First block to process
+    :param last_block: Last block to process
+    :param handlers: List of indexer handlers
+    """
 
-@dataclass
-class BigmapdiffHandlerConfig:
-    callback: str
-    pattern: List[BigmapdiffHandlerPatternConfig]
-
-
-@dataclass
-class BigmapdiffIndexConfig:
-    kind: Literal['bigmapdiff']
-    datasource: Union[str, TzktDatasourceConfig]
-    contract: Union[str, ContractConfig]
-    handlers: List[BigmapdiffHandlerConfig]
+    kind: Literal["operation"]
+    contracts: List[Union[str, ContractConfig]]
+    handlers: List[OperationHandlerConfig]
+    first_block: int = 0
+    last_block: int = 0
 
     @property
-    def tzkt_config(self) -> TzktDatasourceConfig:
-        if not isinstance(self.datasource, TzktDatasourceConfig):
-            raise RuntimeError('Config is not initialized')
-        return self.datasource
+    def contract_configs(self) -> List[ContractConfig]:
+        for contract in self.contracts:
+            if not isinstance(contract, ContractConfig):
+                raise RuntimeError('Config is not initialized')
+        return cast(List[ContractConfig], self.contracts)
 
 
 @dataclass
-class BlockHandlerConfig:
-    callback: str
+class BigMapHandlerPatternConfig:
+    contract: Union[str, ContractConfig]
+    path: str
+
+    def __post_init_post_parse__(self):
+        self._key_type_cls = None
+        self._value_type_cls = None
+
+    @property
+    def contract_config(self) -> ContractConfig:
+        if not isinstance(self.contract, ContractConfig):
+            raise RuntimeError('Config is not initialized')
+        return self.contract
+
+    @property
+    def key_type_cls(self) -> Type:
+        if self._key_type_cls is None:
+            raise RuntimeError('Config is not initialized')
+        return self._key_type_cls
+
+    @key_type_cls.setter
+    def key_type_cls(self, typ: Type) -> None:
+        self._key_type_cls = typ
+
+    @property
+    def value_type_cls(self) -> Type:
+        if self._value_type_cls is None:
+            raise RuntimeError('Config is not initialized')
+        return self._value_type_cls
+
+    @value_type_cls.setter
+    def value_type_cls(self, typ: Type) -> None:
+        self._value_type_cls = typ
+
+
+@dataclass
+class BigMapHandlerConfig(HandlerConfig):
+    pattern: List[BigMapHandlerPatternConfig]
+
+
+@dataclass
+class BigMapIndexConfig(IndexConfig):
+    kind: Literal['big_map']
+    datasource: Union[str, TzktDatasourceConfig]
+    handlers: List[BigMapHandlerConfig]
+    first_block: int = 0
+    last_block: int = 0
+
+
+@dataclass
+class BlockHandlerConfig(HandlerConfig):
     pattern = None
 
 
 @dataclass
-class BlockIndexConfig:
+class BlockIndexConfig(IndexConfig):
     kind: Literal['block']
     datasource: Union[str, TzktDatasourceConfig]
     handlers: List[BlockHandlerConfig]
-
-    @property
-    def tzkt_config(self) -> TzktDatasourceConfig:
-        if not isinstance(self.datasource, TzktDatasourceConfig):
-            raise RuntimeError('Config is not initialized')
-        return self.datasource
+    first_block: int = 0
+    last_block: int = 0
 
 
 @dataclass
 class IndexTemplateConfig:
+    kind = 'template'
     template: str
     values: Dict[str, str]
 
 
-IndexConfigT = Union[OperationIndexConfig, BigmapdiffIndexConfig, BlockIndexConfig, IndexTemplateConfig]
-IndexConfigTemplateT = Union[OperationIndexConfig, BigmapdiffIndexConfig, BlockIndexConfig]
+IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig, IndexTemplateConfig]
+IndexConfigTemplateT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig]
 
 
 @dataclass
@@ -349,7 +369,7 @@ class HasuraConfig:
 class DipDupConfig:
     """Main dapp config
 
-    :param spec_version: Version of specification, always 0.0.1 for now
+    :param spec_version: Version of specification
     :param package: Name of dapp python package, existing or not
     :param contracts: Mapping of contract aliases and contract configs
     :param datasources: Mapping of datasource aliases and datasource configs
@@ -369,10 +389,16 @@ class DipDupConfig:
     hasura: Optional[HasuraConfig] = None
 
     def __post_init_post_parse__(self):
+        if isinstance(self.database, SqliteDatabaseConfig) and self.hasura:
+            raise ConfigurationError('SQLite DB engine is not supported by Hasura')
+
         _logger.info('Substituting index templates')
         for index_name, index_config in self.indexes.items():
             if isinstance(index_config, IndexTemplateConfig):
-                template = self.templates[index_config.template]
+                try:
+                    template = self.templates[index_config.template]
+                except KeyError as e:
+                    raise ConfigurationError(f'Template `{index_config.template}` not found in `templates` config section') from e
                 raw_template = json.dumps(template, default=pydantic_encoder)
                 for key, value in index_config.values.items():
                     value_regex = r'<[ ]*' + key + r'[ ]*>'
@@ -387,16 +413,40 @@ class DipDupConfig:
         for index_config in self.indexes.values():
             if isinstance(index_config, OperationIndexConfig):
                 if isinstance(index_config.datasource, str):
-                    index_config.datasource = self.datasources[index_config.datasource]
-                if isinstance(index_config.contract, str):
-                    index_config.contract = self.contracts[index_config.contract]
+                    try:
+                        index_config.datasource = self.datasources[index_config.datasource]
+                    except KeyError as e:
+                        raise ConfigurationError(f'Datasource `{index_config.datasource}` not found in `datasources` config section') from e
+
+                for i, contract in enumerate(index_config.contracts):
+                    if isinstance(contract, str):
+                        try:
+                            index_config.contracts[i] = self.contracts[contract]
+                        except KeyError as e:
+                            raise ConfigurationError(f'Contract `{contract}` not found in `contracts` config section') from e
 
                 for handler in index_config.handlers:
-                    if isinstance(handler.pattern, list):
-                        callback_patterns[handler.callback].append(handler.pattern)
-                        for pattern in handler.pattern:
-                            if isinstance(pattern.destination, str):
+                    callback_patterns[handler.callback].append(handler.pattern)
+                    for pattern in handler.pattern:
+                        if isinstance(pattern.destination, str):
+                            try:
                                 pattern.destination = self.contracts[pattern.destination]
+                            except KeyError as e:
+                                raise ConfigurationError(f'Contract `{pattern.destination}` not found in `contracts` config section') from e
+
+            elif isinstance(index_config, BigMapIndexConfig):
+                if isinstance(index_config.datasource, str):
+                    index_config.datasource = self.datasources[index_config.datasource]
+
+                for handler in index_config.handlers:
+                    callback_patterns[handler.callback].append(handler.pattern)
+                    for pattern in handler.pattern:
+                        if isinstance(pattern.contract, str):
+                            try:
+                                pattern.contract = self.contracts[pattern.contract]
+                            except KeyError as e:
+                                raise ConfigurationError(f'Contract `{pattern.contract}` not found in `contracts` config section') from e
+
             else:
                 raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
 
@@ -422,80 +472,132 @@ class DipDupConfig:
     @classmethod
     def load(
         cls,
-        filename: str,
+        filenames: List[str],
     ) -> 'DipDupConfig':
 
         current_workdir = os.path.join(os.getcwd())
-        filename = os.path.join(current_workdir, filename)
 
-        _logger.info('Loading config from %s', filename)
-        with open(filename) as file:
-            raw_config = file.read()
+        json_config: Dict[str, Any] = {}
+        for filename in filenames:
+            filename = os.path.join(current_workdir, filename)
 
-        _logger.info('Substituting environment variables')
-        for match in re.finditer(ENV_VARIABLE_REGEX, raw_config):
-            variable, default_value = match.group(1), match.group(2)
-            value = env.get(variable)
-            placeholder = '${' + variable + ':-' + default_value + '}'
-            raw_config = raw_config.replace(placeholder, value or default_value)
+            _logger.info('Loading config from %s', filename)
+            with open(filename) as file:
+                raw_config = file.read()
+
+            _logger.info('Substituting environment variables')
+            for match in re.finditer(ENV_VARIABLE_REGEX, raw_config):
+                variable, default_value = match.group(1), match.group(2)
+                value = env.get(variable)
+                if not default_value and not value:
+                    raise ConfigurationError(f'Environment variable `{variable}` is not set')
+                placeholder = '${' + variable + ':-' + default_value + '}'
+                raw_config = raw_config.replace(placeholder, value or default_value)
+
+            json_config = {
+                **json_config,
+                **YAML(typ='base').load(raw_config),
+            }
 
-        json_config = YAML(typ='base').load(raw_config)
         config = cls(**json_config)
         return config
 
+    async def _initialize_index_state(
+        self, index_name: str, index_config: Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig]
+    ):
+        _logger.info('Getting state for index `%s`', index_name)
+        index_hash = index_config.hash()
+        state = await State.get_or_none(
+            index_name=index_name,
+            index_type=index_config.kind,
+        )
+        if state is None:
+            state = State(
+                index_name=index_name,
+                index_type=index_config.kind,
+                hash=index_hash,
+                level=index_config.first_block - 1,
+            )
+            await state.save()
+
+        elif state.hash != index_hash:
+            _logger.warning('Config hash mismatch, reindexing')
+            await reindex()
+
+        index_config.state = state
+
+    async def _initialize_handler_callback(self, handler_config: HandlerConfig) -> None:
+        _logger.info('Registering handler callback `%s`', handler_config.callback)
+        handler_module = importlib.import_module(f'{self.package}.handlers.{handler_config.callback}')
+        callback_fn = getattr(handler_module, handler_config.callback)
+        handler_config.callback_fn = callback_fn
+
     async def initialize(self) -> None:
         _logger.info('Setting up handlers and types for package `%s`', self.package)
 
-        rollback_fn = getattr(importlib.import_module(f'{self.package}.handlers.{ROLLBACK_HANDLER}'), ROLLBACK_HANDLER)
-
         for index_name, index_config in self.indexes.items():
+
+            if isinstance(index_config, IndexTemplateConfig):
+                raise RuntimeError('Config is not initialized')
+
+            await self._initialize_index_state(index_name, index_config)
+
             if isinstance(index_config, OperationIndexConfig):
-                _logger.info('Getting state for index `%s`', index_name)
-                index_config.rollback_fn = rollback_fn
-                index_hash = index_config.hash()
-                state = await State.get_or_none(
-                    index_name=index_name,
-                    index_type=IndexType.operation,
-                )
-                if state is None:
-                    state = State(
-                        index_name=index_name,
-                        index_type=IndexType.operation,
-                        hash=index_hash,
-                        level=index_config.first_block - 1,
-                    )
-                    await state.save()
 
-                elif state.hash != index_hash:
-                    _logger.warning('Config hash mismatch, reindexing')
-                    await reindex()
+                for operation_handler_config in index_config.handlers:
+                    await self._initialize_handler_callback(operation_handler_config)
 
-                index_config.state = state
-
-                for handler in index_config.handlers:
-                    _logger.info('Registering handler callback `%s`', handler.callback)
-                    handler_module = importlib.import_module(f'{self.package}.handlers.{handler.callback}')
-                    callback_fn = getattr(handler_module, handler.callback)
-                    handler.callback_fn = callback_fn
-
-                    for pattern in handler.pattern:
-                        _logger.info('Registering parameter type for entrypoint `%s`', pattern.entrypoint)
+                    for operation_pattern_config in operation_handler_config.pattern:
+                        _logger.info('Registering parameter type for entrypoint `%s`', operation_pattern_config.entrypoint)
                         parameter_type_module = importlib.import_module(
                             f'{self.package}'
                             f'.types'
-                            f'.{pattern.contract_config.module_name}'
+                            f'.{operation_pattern_config.contract_config.module_name}'
                             f'.parameter'
-                            f'.{camel_to_snake(pattern.entrypoint)}'
+                            f'.{camel_to_snake(operation_pattern_config.entrypoint)}'
                         )
-                        parameter_type_cls = getattr(parameter_type_module, snake_to_camel(pattern.entrypoint))
-                        pattern.parameter_type_cls = parameter_type_cls
+                        parameter_type_cls = getattr(
+                            parameter_type_module, snake_to_camel(operation_pattern_config.entrypoint) + 'Parameter'
+                        )
+                        operation_pattern_config.parameter_type_cls = parameter_type_cls
 
                         _logger.info('Registering storage type')
                         storage_type_module = importlib.import_module(
-                            f'{self.package}' f'.types' f'.{pattern.contract_config.module_name}' f'.storage'
+                            f'{self.package}.types.{operation_pattern_config.contract_config.module_name}.storage'
                         )
-                        storage_type_cls = getattr(storage_type_module, 'Storage')
-                        pattern.storage_type_cls = storage_type_cls
+                        storage_type_cls = getattr(
+                            storage_type_module, snake_to_camel(operation_pattern_config.contract_config.module_name) + 'Storage'
+                        )
+                        operation_pattern_config.storage_type_cls = storage_type_cls
+
+            elif isinstance(index_config, BigMapIndexConfig):
+                for big_map_handler_config in index_config.handlers:
+                    await self._initialize_handler_callback(big_map_handler_config)
+
+                    for big_map_pattern_config in big_map_handler_config.pattern:
+                        _logger.info('Registering big map types for path `%s`', big_map_pattern_config.path)
+                        key_type_module = importlib.import_module(
+                            f'{self.package}'
+                            f'.types'
+                            f'.{big_map_pattern_config.contract_config.module_name}'
+                            f'.big_map'
+                            f'.{camel_to_snake(big_map_pattern_config.path)}_key'
+                        )
+                        key_type_cls = getattr(key_type_module, snake_to_camel(big_map_pattern_config.path + '_key'))
+                        big_map_pattern_config.key_type_cls = key_type_cls
+
+                        value_type_module = importlib.import_module(
+                            f'{self.package}'
+                            f'.types'
+                            f'.{big_map_pattern_config.contract_config.module_name}'
+                            f'.big_map'
+                            f'.{camel_to_snake(big_map_pattern_config.path)}_value'
+                        )
+                        value_type_cls = getattr(value_type_module, snake_to_camel(big_map_pattern_config.path + '_value'))
+                        big_map_pattern_config.value_type_cls = value_type_cls
+
+            else:
+                raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
 
 
 @dataclass
