@@ -8,8 +8,10 @@ from pydantic import BaseModel
 from pydantic.dataclasses import dataclass
 from tortoise import Model, fields
 
-ParameterType = TypeVar('ParameterType')
+ParameterType = TypeVar('ParameterType', bound=BaseModel)
 StorageType = TypeVar('StorageType', bound=BaseModel)
+KeyType = TypeVar('KeyType', bound=BaseModel)
+ValueType = TypeVar('ValueType', bound=BaseModel)
 
 
 _logger = logging.getLogger(__name__)
@@ -17,7 +19,7 @@ _logger = logging.getLogger(__name__)
 
 class IndexType(Enum):
     operation = 'operation'
-    bigmapdiff = 'bigmapdiff'
+    big_map = 'big_map'
     block = 'block'
     schema = 'schema'
 
@@ -42,17 +44,11 @@ class OperationData:
     hash: str
     counter: int
     sender_address: str
-    gas_limit: int
-    gas_used: int
-    storage_limit: int
-    storage_used: int
-    baker_fee: int
-    storage_fee: int
-    allocation_fee: int
     target_address: str
     amount: int
     status: str
     has_internals: bool
+    storage: Dict[str, Any]
     block: Optional[str] = None
     sender_alias: Optional[str] = None
     nonce: Optional[int] = None
@@ -61,7 +57,6 @@ class OperationData:
     parameter_json: Optional[Any] = None
     initiator_address: Optional[str] = None
     parameter: Optional[str] = None
-    storage: Optional[Dict[str, Any]] = None
     diffs: Optional[List[Dict[str, Any]]] = None
 
     def _merge_bigmapdiffs(self, storage_dict: Dict[str, Any], bigmap_name: str, array: bool) -> None:
@@ -82,17 +77,18 @@ class OperationData:
     def _process_storage(self, storage_type: Type[StorageType], storage: Dict, prefix: str = None):
         for key, field in storage_type.__fields__.items():
 
-            bigmap_name = key if prefix is None else '.'.join([prefix, key])
-
             if key == '__root__':
                 continue
 
             if field.alias:
                 key = field.alias
 
+            bigmap_name = key if prefix is None else '.'.join([prefix, key])
+
             # NOTE: TzKT could return bigmaps as object or as array of key-value objects. We need to guess this from storage.
             # TODO: This code should be a part of datasource module.
             if field.type_ not in (int, bool) and isinstance(storage[key], int):
+                _logger.debug(field.type_)
                 if hasattr(field.type_, '__fields__') and 'key' in field.type_.__fields__ and 'value' in field.type_.__fields__:
                     storage[key] = []
                     if self.diffs:
@@ -123,13 +119,45 @@ class OperationData:
 
 
 @dataclass
-class OperationContext(Generic[ParameterType]):
+class OperationContext(Generic[ParameterType, StorageType]):
     data: OperationData
     parameter: ParameterType
-    storage: Optional[Any] = None
+    storage: StorageType
+
+
+class BigMapAction(Enum):
+    ADD = 'add_key'
+    UPDATE = 'update_key'
+    REMOVE = 'remove_key'
 
 
 @dataclass
-class HandlerContext:
+class BigMapContext(Generic[KeyType, ValueType]):
+    action: BigMapAction
+    key: KeyType
+    value: Optional[ValueType]
+
+
+@dataclass
+class BigMapData:
+    id: int
+    level: int
+    operation_id: int
+    timestamp: datetime
+    bigmap: int
+    contract_address: str
+    path: str
+    action: str
+    key: Optional[Any] = None
+    value: Optional[Any] = None
+
+
+@dataclass
+class OperationHandlerContext:
     operations: List[OperationData]
     template_values: Optional[Dict[str, str]]
+
+
+@dataclass
+class BigMapHandlerContext:
+    template_values: Optional[Dict[str, str]]
